#!/usr/bin/env python

# "Go Big" Program Launcher

COPYRIGHT = "Copyright 2009 Shane Harper" # http://shaneharper.net

LICENSE = """\
This program is free software: you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation, either version 3 of the License, or \
(at your option) any later version.

This program is distributed in the hope that it will be useful, \
but WITHOUT ANY WARRANTY; without even the implied warranty of \
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \
GNU General Public License for more details.

You should have received a copy of the GNU General Public License \
along with this program.  If not, see <http://www.gnu.org/licenses/>.\
"""


import pygtk
pygtk.require('2.0')
import gobject
import gmenu
import gtk
import os
import pango
import shutil
import subprocess
import sys


APPLICATIONS_PER_ROW = 7
# if DEFAULT_FONT_SIZE_OVERRIDE is None:  Use default font
# if DEFAULT_FONT_SIZE_OVERRIDE > 0:      Override default font size
# if DEFAULT_FONT_SIZE_OVERRIDE == 0:     Application names aren't displayed - only the icons are displayed.
DEFAULT_FONT_SIZE_OVERRIDE = None
CATEGORY_ICON_SIZE_IN_PIXELS = 32
SORT_APPS_ALPHABETICALLY_WITHIN_A_CATEGORY = True
CATEGORY_TITLE_INDENT = 5
WIN_BORDER_WIDTH = 5
APP_WIDGET_BORDER_WIDTH = 5
SCROLLBAR_WIDTH = 18
COLUMN_GAP_APP_RATIO = 0.6
APPLICATION_WIDGET_WIDTH = int((gtk.gdk.screen_width() - 2*WIN_BORDER_WIDTH - SCROLLBAR_WIDTH) / (2*APPLICATIONS_PER_ROW + COLUMN_GAP_APP_RATIO))
COLUMN_GAP = int(COLUMN_GAP_APP_RATIO * APPLICATION_WIDGET_WIDTH)
SELECTED_APP_ICON_SIZE_IN_PIXELS = int(0.82 * (APPLICATION_WIDGET_WIDTH - 2*APP_WIDGET_BORDER_WIDTH))
APP_ICON_SIZE_IN_PIXELS = int(0.85 * SELECTED_APP_ICON_SIZE_IN_PIXELS)
VERTICAL_GAP_BETWEEN_APPS = int(0.06 * APPLICATION_WIDGET_WIDTH)
VERTICAL_GAP_BETWEEN_CATEGORIES = int(0.3 * APPLICATION_WIDGET_WIDTH)


_icon_theme = gtk.icon_theme_get_default()

_icon_cache_folder = os.path.join(
    os.environ['XDG_CACHE_HOME'] if os.environ.has_key('XDG_CACHE_HOME') and "" != os.environ['XDG_CACHE_HOME']
        else os.path.join(os.environ['HOME'], '.cache'),
    'go_big')
if not os.path.isdir(_icon_cache_folder):
    os.makedirs(_icon_cache_folder)

def clear_icon_cache():
    for filename in os.listdir(_icon_cache_folder):
        os.remove(os.path.join(_icon_cache_folder, filename))

def get_icon(item, size):
    if item.get_icon() is None:
        return None

    cache_file_path = os.path.join(_icon_cache_folder,
                                   (item.get_icon() + '--' + str(size) + '.png').replace('/', '_'))
    try:
        return gtk.gdk.pixbuf_new_from_file(cache_file_path)
    except:
        pixbuf = get_icon__no_cache(item, size)
        try:
            if pixbuf:
                pixbuf.save(cache_file_path, 'png')
        except Exception, error:
            print "Failed to save icon in cache. Pathname:", cache_file_path, "  Error:", error
        return pixbuf

# The following function is based on util.getIcon from Alacarte (http://en.wikipedia.org/wiki/Alacarte).
def get_icon__no_cache(item, size):
    if item == None:
        return None
    if isinstance(item, str):
        iconName = item
    else:
        iconName = item.get_icon()
    if iconName and not '/' in iconName and iconName[-3:] in ('png', 'svg', 'xpm'):
        iconName = iconName[:-4]
    pixbuf = None
    try:
        pixbuf = _icon_theme.load_icon(iconName, size, 0)
    except:
        if iconName and '/' in iconName:
            try:
                pixbuf = gtk.gdk.pixbuf_new_from_file_at_size(iconName, size, size)
            except:
                pass
        if pixbuf == None:
            if item.get_type() == gmenu.TYPE_DIRECTORY:
                iconName = 'gnome-fs-directory'
            elif item.get_type() == gmenu.TYPE_ENTRY:
                iconName = 'application-default-icon'
            try:
                pixbuf = _icon_theme.load_icon(iconName, size, 0)
            except:
                return None
    if pixbuf == None:
        return None
    if pixbuf.get_width() != size and pixbuf.get_height() != size:
        # preserve aspect ratio
        aspect_ratio = float(pixbuf.get_width())/pixbuf.get_height()
        pixbuf = pixbuf.scale_simple(size if aspect_ratio >= 1 else int(size * aspect_ratio),
                                     size if aspect_ratio <= 1 else int(size / aspect_ratio),
                                     gtk.gdk.INTERP_HYPER)
    return pixbuf


def brighten_pixbuf(pixbuf, factor = 1.1):
    pixels = pixbuf.get_pixels_array()
    for y in range(pixbuf.get_height()):
        for x in range(pixbuf.get_width()):
            pixels[x,y,0] = min(255, int(pixels[x,y,0]) * factor)
            pixels[x,y,1] = min(255, int(pixels[x,y,1]) * factor)
            pixels[x,y,2] = min(255, int(pixels[x,y,2]) * factor)


def get_user_path(folder):
    path = os.path.join(os.environ['XDG_DATA_HOME'], folder) if os.environ.has_key('XDG_DATA_HOME') and "" != os.environ['XDG_DATA_HOME'] \
        else os.path.join(os.environ['HOME'], '.local', 'share', folder)
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


def create_gtk_menu_item(label, on_activate_function, use_underline = True):
    menu_item = gtk.MenuItem(label, use_underline)
    menu_item.connect("activate", on_activate_function)
    return menu_item


def get_all_menu_items(directory):
    """ Return a list of all menu items including items in submenus.

    The items won't necessarily be sorted alphabetically by name: The order of
    the items is exactly as read from 'directory'.
    """
    menu_items = []
    for item in directory.get_contents():
        if gmenu.TYPE_ENTRY == item.get_type():
            menu_items += [item]
        elif gmenu.TYPE_DIRECTORY == item.get_type():
            menu_items += get_all_menu_items(item)
    return menu_items


class ApplicationWidget:
    def __init__(self, x, y, height, item, pango_layout):
        self.x, self.y = x, y
        self.height = height
        self.item = item
        self.icon = get_icon(item, APP_ICON_SIZE_IN_PIXELS)
        self.pango_layout = pango_layout


class CategoryHeading:
    def __init__(self, x, y, elem):
        self.x, self.y = x, y
        self.name = elem.get_name()
        self.desktop_file_path = elem.get_desktop_file_path()
        self.icon = get_icon(elem, CATEGORY_ICON_SIZE_IN_PIXELS)


class ProgramLauncher:
    def __init__(self, menu="applications.menu"):
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_decorated(False)
        self.window.connect("destroy", lambda w: gtk.main_quit())
        self.window.maximize()
        self.drawing_area = gtk.DrawingArea()
        self.drawing_area.set_flags(gtk.CAN_FOCUS)
        self.drawing_area.grab_focus()  # We need focus to receive key press events.
        self.drawing_area.set_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.KEY_PRESS_MASK | gtk.gdk.LEAVE_NOTIFY_MASK | gtk.gdk.POINTER_MOTION_MASK)
        self.drawing_area.connect('button-press-event', self._on_button_press_event)
        self.drawing_area.connect('button-release-event', self._on_button_release_event)
        self.drawing_area.connect("drag_data_get", self._drag_data_get)
        self.drawing_area.connect("expose-event", self._on_expose_event)
        self.drawing_area.connect('key-press-event', self._on_key_press_event)
        self.drawing_area.connect("leave-notify-event", self._on_leave_notify_event)
        self.drawing_area.connect('motion-notify-event', self._on_motion_notify_event)
        self.drawing_area.connect("popup-menu", self._on_popup_menu)
        self.title_pango_layout = self.drawing_area.create_pango_layout("")
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_policy(hscrollbar_policy = gtk.POLICY_NEVER, vscrollbar_policy = gtk.POLICY_AUTOMATIC)
        scrolled_window.add_with_viewport(self.drawing_area)
        self.window.add(scrolled_window)
        self.window.show_all()

        self.application_menu = gtk.Menu()
        self.application_menu.append(create_gtk_menu_item("Edit", self._on_edit_application))
        self.application_menu.show_all()

        self.menu_tree = gmenu.lookup_tree(menu)
        if self.menu_tree.get_menu_file() is None: raise Exception("gmenu.lookup_tree failed.")
        self.menu_tree.add_monitor(self._on_menu_changed, None)
        self.do_layout_idle_handler_id = None

        _icon_theme.connect('changed', lambda icon_theme: self.refresh())

        self.do_layout()

    def do_layout(self):
        """ Sets the position of all icons and text and then requests that the window be drawn. """

        def layout_category_apps(applications, y):
            if SORT_APPS_ALPHABETICALLY_WITHIN_A_CATEGORY:
                applications.sort(key=lambda e: e.get_name().lower())
            for row_idx in range(0, len(applications), APPLICATIONS_PER_ROW):
                row_applications = []
                x = WIN_BORDER_WIDTH
                for app in applications[row_idx : row_idx + APPLICATIONS_PER_ROW]:
                    pango_layout = self.drawing_area.create_pango_layout(app.get_name())
                    if DEFAULT_FONT_SIZE_OVERRIDE is not None:
                        font = pango_layout.get_context().get_font_description()
                        font.set_size(DEFAULT_FONT_SIZE_OVERRIDE*pango.SCALE)
                        pango_layout.set_font_description(font)
                    pango_layout.set_width((APPLICATION_WIDGET_WIDTH - APP_WIDGET_BORDER_WIDTH*2) * pango.SCALE)
                    pango_layout.set_wrap(pango.WRAP_WORD_CHAR)
                    pango_layout.set_alignment(pango.ALIGN_CENTER)
                    row_applications += [(x, y, app, pango_layout)]
                    x += APPLICATION_WIDGET_WIDTH
                # set height for this row
                row_height = max([pango_layout.get_pixel_size()[1] for _, _, _, pango_layout in row_applications]) \
                             + SELECTED_APP_ICON_SIZE_IN_PIXELS + 2*APP_WIDGET_BORDER_WIDTH + VERTICAL_GAP_BETWEEN_APPS
                self.applications += [ApplicationWidget(app_x, app_y, row_height, app, pango_layout)
                                      for app_x, app_y, app, pango_layout in row_applications]
                y += row_height
            return y

        self.category_headings = []
        self.applications = []

        y = layout_category_apps(filter(lambda item: gmenu.TYPE_ENTRY == item.get_type(),
                                        self.menu_tree.get_root_directory().get_contents()),
                                 WIN_BORDER_WIDTH)

        for elem in self.menu_tree.get_root_directory().get_contents():
            if gmenu.TYPE_DIRECTORY == elem.get_type():
                if y > WIN_BORDER_WIDTH:
                    y += VERTICAL_GAP_BETWEEN_CATEGORIES
                self.category_headings += [CategoryHeading(WIN_BORDER_WIDTH, y, elem)]
                y += CATEGORY_ICON_SIZE_IN_PIXELS + 16
                y = layout_category_apps(get_all_menu_items(elem), y)

        # Layout icons in two columns.
        # Icons below split_y will be put in the second column.
        height, split_y = min((max(split_y - VERTICAL_GAP_BETWEEN_CATEGORIES, y - split_y + WIN_BORDER_WIDTH), split_y)
                               for split_y in ([cat.y for cat in self.category_headings] + [y]))

        for item in self.category_headings + self.applications:
            if item.y >= split_y:
                item.y += WIN_BORDER_WIDTH - split_y
                item.x += APPLICATIONS_PER_ROW*APPLICATION_WIDGET_WIDTH + COLUMN_GAP

        # Set drawing_area height so that vertical scrollbar range is correct.
        self.drawing_area.set_size_request(1, height + WIN_BORDER_WIDTH)

        self.drawing_area.window.invalidate_rect(gtk.gdk.Rectangle(0, 0, *self.drawing_area.window.get_size()), True)

    def _draw_centered_icon(self, gc, pixbuf, center_x, center_y):
        if pixbuf is not None:
            self.drawing_area.window.draw_pixbuf(gc, pixbuf, 0, 0,
                                                 center_x - pixbuf.get_width()/2,
                                                 center_y - pixbuf.get_height()/2)

    def _on_expose_event(self, area, event):
        try: self.selected_widget
        except AttributeError:
            # This occurs the first time this function is called. self._set_selected_widget() isn't
            # called from self.__init__() because calling self._set_selected_widget() before the window
            # is mapped can result in the wrong app being selected (because screen co-ordinates are
            # used instead of window co-ordinates).
            self._set_selected_widget()

        style = self.drawing_area.get_style()
        normal_gc = style.fg_gc[gtk.STATE_NORMAL]

        # Draw category titles
        for cat in self.category_headings:
            if cat is self.selected_widget:
                icon = cat.icon.copy()
                brighten_pixbuf(icon, 1.1)
            else:
                icon = cat.icon
            self._draw_centered_icon(normal_gc, icon, cat.x + CATEGORY_TITLE_INDENT + CATEGORY_ICON_SIZE_IN_PIXELS/2,
                                     cat.y + CATEGORY_ICON_SIZE_IN_PIXELS/2)
            self.title_pango_layout.set_text(cat.name)
            self.drawing_area.window.draw_layout(normal_gc, cat.x + CATEGORY_ICON_SIZE_IN_PIXELS + CATEGORY_TITLE_INDENT + 15,
                                                 cat.y + (CATEGORY_ICON_SIZE_IN_PIXELS - self.title_pango_layout.get_pixel_size()[1])/2,
                                                 self.title_pango_layout)
            self.drawing_area.window.draw_line(style.fg_gc[gtk.STATE_INSENSITIVE], cat.x,
                                               cat.y + CATEGORY_ICON_SIZE_IN_PIXELS + 6,
                                               cat.x + APPLICATIONS_PER_ROW*APPLICATION_WIDGET_WIDTH,
                                               cat.y + CATEGORY_ICON_SIZE_IN_PIXELS + 6)

        # Draw application widgets
        for app in self.applications:
            state = gtk.STATE_SELECTED if app is self.selected_widget else gtk.STATE_NORMAL
            if gtk.STATE_SELECTED == state:
                self.drawing_area.window.draw_rectangle(style.bg_gc[state], True, app.x, app.y, APPLICATION_WIDGET_WIDTH, app.height)
            self._draw_centered_icon(normal_gc,
                                     get_icon(app.item, SELECTED_APP_ICON_SIZE_IN_PIXELS) if gtk.STATE_SELECTED == state else app.icon,
                                     app.x + APPLICATION_WIDGET_WIDTH/2,
                                     app.y + APP_WIDGET_BORDER_WIDTH + SELECTED_APP_ICON_SIZE_IN_PIXELS/2)
            self.drawing_area.window.draw_layout(style.fg_gc[state], app.x + APP_WIDGET_BORDER_WIDTH,
                                                 app.y + SELECTED_APP_ICON_SIZE_IN_PIXELS + APP_WIDGET_BORDER_WIDTH + 1, app.pango_layout)

        return True

    def display_error(self, line1, line2 = "", title = "Error"):
        error_dialog = gtk.MessageDialog(parent=self.window, flags=gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                                         type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_OK, message_format=line1)
        error_dialog.set_title(title)
        error_dialog.format_secondary_text(line2)
        error_dialog.run()
        error_dialog.destroy()

    def _start_selected_application(self):
        if not isinstance(self.selected_widget, ApplicationWidget):
            return False

        cmd_and_args = ['gnome-terminal', '--execute'] if self.selected_widget.item.get_launch_in_terminal() else []
        # Replace special arguments. (See http://standards.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html)
        for a in self.selected_widget.item.get_exec().split():
            if '%c' == a:
                cmd_and_args += [self.selected_widget.item.get_name()]
            elif '%i' == a:
                icon = self.selected_widget.item.get_icon()
                if icon:
                    cmd_and_args += ['--icon='+icon]
            elif '%k' == a:
                cmd_and_args += [self.selected_widget.item.get_desktop_file_path()]
            elif a not in ['%f', '%F', '%u', '%U']:
                cmd_and_args += [a]

        try:
            if len(cmd_and_args) == 0:
                raise Exception("No command.")
            subprocess.Popen(cmd_and_args)
        except Exception, error:
            self.display_error("Could not launch %s" % self.selected_widget.item.get_name(), str(error))
            return False
        return True

    def _on_button_press_event(self, widget, event):
        self.drag = False
        if event.button == 1:
            self.button_press_x, self.button_press_y = int(event.x), int(event.y)
            self.button_press_event = event
            return True
        return False

    def _on_button_release_event(self, widget, event):
        if event.button == 1 and not self.drag:
            # XXX Could we have missed a motion_notify_event??? (In which case self.selected_widget may now be incorrect)
            if self._start_selected_application():
                gtk.main_quit()
            return True
        elif event.button == 3:
            if isinstance(self.selected_widget, ApplicationWidget):
                self.application_menu.popup(None, None, None, event.button, event.get_time())
                return True
            elif isinstance(self.selected_widget, CategoryHeading):
                menu = gtk.Menu()
                menu.append(create_gtk_menu_item("Edit", lambda widget: self._on_edit_category(widget, self.selected_widget)))
                menu.show_all()
                menu.popup(None, None, None, event.button, event.get_time())
                return True
        return False

    def refresh(self):
        clear_icon_cache()
        self.do_layout()
        self._set_selected_widget()

    def _on_key_press_event(self, widget, event):
        if event.keyval == gtk.keysyms.Escape:
            gtk.main_quit()
            return True
        elif event.keyval == gtk.keysyms.F1:
            about_dialog = gtk.AboutDialog()
            about_dialog.set_name("Go Big")
            about_dialog.set_website("http://shaneharper.net/utils/go_big")
#            about_dialog.set_version(XXX)
            about_dialog.set_copyright(COPYRIGHT)
            about_dialog.set_license(LICENSE)
            about_dialog.set_wrap_license(True)
#            about_dialog.set_authors(["Shane Harper"])
            about_dialog.run()
            about_dialog.destroy()
            return True
        elif event.keyval == gtk.keysyms.F5:  # XXX This wouldn't be necessary if the validity of the icon cache was ensured at start-up!
            self.refresh()
            return True
        elif event.keyval == gtk.keysyms.Return:
            if self._start_selected_application():
                gtk.main_quit()
            return True
        return False

    def _redraw_selected_widget(self):
        if isinstance(self.selected_widget, ApplicationWidget):
            self.drawing_area.window.invalidate_rect(gtk.gdk.Rectangle(self.selected_widget.x, self.selected_widget.y, APPLICATION_WIDGET_WIDTH, self.selected_widget.height), True)
        elif isinstance(self.selected_widget, CategoryHeading):
            self.drawing_area.window.invalidate_rect(gtk.gdk.Rectangle(self.selected_widget.x+CATEGORY_TITLE_INDENT, self.selected_widget.y, CATEGORY_ICON_SIZE_IN_PIXELS, CATEGORY_ICON_SIZE_IN_PIXELS), True)

    def _find_selected_widget(self, x, y):
        for app in self.applications:
            if 0 <= x-app.x < APPLICATION_WIDGET_WIDTH and 0 <= y-app.y < app.height:
                return app
        for cat in self.category_headings:
            if 0 <= x-(cat.x+CATEGORY_TITLE_INDENT) < CATEGORY_ICON_SIZE_IN_PIXELS and 0 <= y-cat.y < CATEGORY_ICON_SIZE_IN_PIXELS:
                return cat

    def _set_selected_widget(self):
        # XXX if mouse_pointer_is_not_in_our_window: self.selected_widget = None; else:
        self.selected_widget = self._find_selected_widget(*(self.drawing_area.get_pointer()[:2]))

    URI_DRAG_ID = 100

    def _on_motion_notify_event(self, widget, event):
        if event.state & gtk.gdk.BUTTON1_MASK:
            if self.drawing_area.drag_check_threshold(int(self.button_press_x), int(self.button_press_y), int(event.x), int(event.y)):
                self.drag = True
                if isinstance(self.selected_widget, ApplicationWidget):
# XXX               self.drawing_area.drag_source_set_icon_pixbuf(get_icon(self.selected_widget.item, 24))
                    self.drawing_area.drag_begin([("text/uri-list", 0, self.URI_DRAG_ID)],
                                                 gtk.gdk.ACTION_LINK | gtk.gdk.ACTION_COPY, 1, self.button_press_event)
            return True

        selected_widget = self._find_selected_widget(event.x, event.y)
        if selected_widget is not self.selected_widget:
            self._redraw_selected_widget()  # Clear previous highlight
            self.selected_widget = selected_widget
            self._redraw_selected_widget()
        return True

    def _drag_data_get(self, widget, context, selection, target_type, event_time):
        if self.URI_DRAG_ID == target_type:
            selection.set_uris(["file://" + self.selected_widget.item.get_desktop_file_path()])

    def _on_popup_menu(self, widget):
        if isinstance(self.selected_widget, ApplicationWidget):
            self.application_menu.popup(None, None, None, 0, 0)
            return True
        return False

    def start_program(self, cmd_and_args, **popen_kwargs):
        try:
            subprocess.Popen(cmd_and_args, **popen_kwargs)
        except Exception, error:
            self.display_error("Could not start %s" % cmd_and_args[0], str(error))

    def _on_edit_application(self, widget):
        users_entry_path = os.path.join(get_user_path('applications'), self.selected_widget.item.get_desktop_file_id())
        if not os.path.isfile(users_entry_path):
            shutil.copy(src=self.selected_widget.item.get_desktop_file_path(), dst=users_entry_path)
        self.start_program(['gnome-desktop-item-edit', users_entry_path], env=os.environ)

    def _on_edit_category(self, widget, category):
        users_entry_path = os.path.join(get_user_path('desktop-directories'), os.path.basename(category.desktop_file_path))
        if not os.path.isfile(users_entry_path):
            shutil.copy(src=category.desktop_file_path, dst=users_entry_path)
        self.start_program(['gnome-desktop-item-edit', users_entry_path], env=os.environ)

    def _on_menu_changed(self, *args):
        if True:
            if self.do_layout_idle_handler_id is None:
                self.do_layout_idle_handler_id = gobject.idle_add(self.do_layout_when_idle)
        else:
            self.do_layout_when_idle()  # XXX Slow: Window typically unresponsive for some time after a menu is changed. It's strange that this function (_on_menu_changed) gets called more often when this statement is executed as opposed to the previous statement that sets up an idle handler.

    def do_layout_when_idle(self):
        self.do_layout()
        self.do_layout_idle_handler_id = None
        self._set_selected_widget()

    def _on_leave_notify_event(self, widget, event):
        if not event.state & (gtk.gdk.BUTTON1_MASK  # Strange that this function gets called on a button1 press!
                             | gtk.gdk.BUTTON3_MASK):  # (Don't clear the highlight before displaying a context menu.)
            self._redraw_selected_widget()
            self.selected_widget = None


if __name__ == "__main__":
    def usage():
        print "Usage: go_big.py [--menu=MENU]"
        sys.exit(1)

    import getopt
    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["menu="])
    except getopt.GetoptError:
       usage()
    menu = "applications.menu"
    for opt, arg in opts:
        if opt == "--menu":
            menu = arg

    ProgramLauncher(menu)
    gtk.main()
